@@ -4,7 +4,7 @@
 
 ;; Author: Adam Porter <adam@alphapapa.net>
 ;; URL: http://github.com/alphapapa/magit-todos
-;; Version: 1.7.2
+;; Version: 1.8
 ;; Package-Requires: ((emacs "26.1") (async "1.9.2") (dash "2.13.0") (f "0.17.2") (hl-todo "1.9.0") (magit "2.13.0") (pcre2el "1.8") (s "1.12.0") (transient "0.2.0"))
 ;; Keywords: magit, vc
 
@@ -148,6 +148,9 @@ A time value as returned by `current-time'.")
 (defvar-local magit-todos-item-cache nil
   "Items found by most recent scan.")
 
+(defvar-local magit-todos-branch-item-cache nil
+  "Items found by most recent branch scan.")
+
 (defvar magit-todos-scanners nil
   "Scanners defined by `magit-todos-defscanner'.")
 
@@ -204,11 +207,19 @@ items."
                  (integer :tag "Automatically, but cache items for N seconds")
                  (const :tag "Manually" nil)))
 
+(defcustom magit-todos-update-remote nil
+  "Automatically scan in remote repositories.
+If nil, remote repos are only scanned when a manual scan is
+initiated (see command `magit-todos-update').  Remote repos
+include ones accessed via TRAMP (i.e. any path for which
+`file-remote-p' returns non-nil)."
+  :type 'boolean)
+
 (defcustom magit-todos-fontify-keyword-headers t
   "Apply keyword faces to group keyword headers."
   :type 'boolean)
 
-(defcustom magit-todos-keyword-suffix (rx (optional "(" (1+ (not (any ")"))) ")") ":")
+(defcustom magit-todos-keyword-suffix (rx (optional (or "(" "[") (1+ (not (any ")" "]"))) (or ")" "]")) ":")
   "Regular expression matching suffixes after keywords.
 e.g. to match a keyword like \"TODO(user):\", use \"([^)]+):\".
 
@@ -218,8 +229,10 @@ account for optional whitespace after the suffix, as this is done
 automatically.
 
 Note: the suffix applies only to non-Org files."
-  :type `(choice (const :tag "Optional username in parens, then required colon (matching e.g. \"TODO:\" or \"TODO(user):\")"
+  :type `(choice (const :tag "Optional suffix in parens, then required colon (matching e.g. \"TODO:\" or \"TODO(foo):\")"
                         ,(rx (optional "(" (1+ (not (any ")"))) ")") ":"))
+                 (const :tag "Optional suffix in parens or brackets, then required colon (matching e.g. \"TODO:\" or \"TODO(foo):\" or \"TODO[foo]:\")"
+                        ,(rx (optional (or "(" "[") (1+ (not (any ")" "]"))) (or ")" "]")) ":"))
                  (const :tag "Required colon (matching e.g. \"TODO:\"" ":")
                  (string :tag "Custom regexp"))
   :package-version '(magit-todos . "1.2"))
@@ -333,28 +346,6 @@ as desired when the built-in scanner is used."
                   (const :tag "Issues (Forge)" issues)
                   (symbol :tag "Specified section"))))
 
-(defcustom magit-todos-insert-at 'bottom
-  "Insert the to-dos section after this section in the Magit status buffer.
-Specific sections may be chosen, using the first symbol returned
-by evaluating \"(magit-section-ident (magit-current-section))\"
-in the status buffer with point on the desired section,
-e.g. `recent' for the \"Recent commits\" section.  Note that this
-may not work exactly as desired when the built-in scanner is
-used."
-  :type '(choice (const :tag "Top" top)
-                 (const :tag "Bottom" bottom)
-                 (const :tag "After untracked files" untracked)
-                 (const :tag "After unstaged files" unstaged)
-                 (symbol :tag "After selected section"))
-  :set (lambda (option value)
-         ;; For convenience, we set the new option with the appropriate value (but,
-         ;; of course, this won't work for users who set it directly with `setq'.)
-         ;; TODO: Remove this option in 1.8.
-         (ignore option)
-         (custom-set-variables `(magit-todos-insert-after ',(list value)
-                                                          'now nil "Changed by setter of obsolete option `magit-todos-insert-at'"))))
-(make-obsolete-variable 'magit-todos-insert-at 'magit-todos-insert-after "1.6")
-
 (defcustom magit-todos-exclude-globs '(".git/")
   "Glob patterns to exclude from searches."
   :type '(repeat string))
@@ -438,6 +429,7 @@ Only necessary when option `magit-todos-update' is nil."
 (defun magit-todos-branch-list-toggle ()
   "Toggle branch diff to-do list in current Magit buffer."
   (interactive)
+  ;; FIXME: If `magit-todos-branch-list' is not just nil/t, the value is lost.
   (setq-local magit-todos-branch-list (not magit-todos-branch-list))
   (magit-todos-update))
 
@@ -505,8 +497,9 @@ or `default-directory' if nil."
                             (file-name-as-directory (expand-file-name directory))
                           default-directory))
         (magit-todos-list-internal directory))
-    ('magit-outside-git-repo (cl-letf (((symbol-function 'magit-toplevel) (lambda (&rest _) default-directory)))
-                               (call-interactively #'magit-todos-list)))))
+    (magit-outside-git-repo
+     (cl-letf (((symbol-function 'magit-toplevel) (lambda (&rest _) default-directory)))
+       (call-interactively #'magit-todos-list)))))
 
 (put 'magit-todos-list 'interactive-only 'magit-todos-list-internal)
 
@@ -601,9 +594,8 @@ To be called in status buffers' `kill-buffer-hook'."
   "Return function to call to scan for items with.
 Chooses automatically in order defined in `magit-todos-scanners'."
   (cl-loop for scanner in magit-todos-scanners
-           ;; I guess it would be better to avoid `eval', but it seems like the natural
-           ;; way to do this.
-           when (eval (alist-get 'test scanner))
+           for availablep = (alist-get 'availablep scanner)
+           when (and availablep (funcall availablep))
            return (alist-get 'function scanner)))
 
 (cl-defun magit-todos--scan-callback (&key callback magit-status-buffer results-regexp process &allow-other-keys)
@@ -620,16 +612,7 @@ buffer for RESULTS-REGEXP."
     (save-excursion
       (goto-char (point-min))
       (while (not (eobp))
-        (--when-let (condition-case err
-                        (magit-todos--line-item results-regexp)
-                      ;; Files with very, very long lines may cause Emacs's regexp matcher to overflow.
-                      ;; Rather than abort the whole scan and raise an error, try to handle it gracefully.
-                      ;; FIXME: This may raise multiple warnings per file.
-                      (error (if (string= "Stack overflow in regexp matcher" (error-message-string err))
-                                 (let ((filename (buffer-substring (point) (1- (re-search-forward ":")))))
-                                   (display-warning 'magit-todos (concat "File has lines too long for Emacs to search.  Consider excluding it from scans: " filename))
-                                   nil)
-                               (signal (car err) (cdr err)))))
+        (--when-let (magit-todos--line-item results-regexp)
           (push it items))
         (forward-line 1)))
     (nreverse items)))
@@ -686,6 +669,7 @@ filenames to be excluded."
                       (push item items)))
                   (cl-incf line-number))))))
         (let ((magit-todos-section-heading heading))
+          (setf (buffer-local-value 'magit-todos-branch-item-cache magit-status-buffer) items)
           (magit-todos--insert-items magit-status-buffer items :branch-p t))))))
 
 (defun magit-todos--delete-section (condition)
@@ -734,46 +718,66 @@ Assumes current buffer is ITEM's buffer."
   "Insert to-do items into current buffer.
 This function should be called from inside a ‘magit-status’ buffer."
   (declare (indent defun))
-  (when magit-todos-active-scan
-    ;; Avoid running multiple scans for a single magit-status buffer.
-    (let ((buffer (process-buffer magit-todos-active-scan)))
-      (when (process-live-p magit-todos-active-scan)
-        (delete-process magit-todos-active-scan))
-      (when (buffer-live-p buffer)
-        (kill-buffer buffer)))
-    (setq magit-todos-active-scan nil))
-  (pcase magit-todos-update
-    ((or 't  ; Automatic
-         ;; Manual and updating now
-         (and 'nil (guard magit-todos-updating))
-         ;; Caching and cache expired
-         (and (pred integerp) (guard (or magit-todos-updating  ; Forced update
-                                         (>= (float-time
-                                              (time-subtract (current-time)
-                                                             magit-todos-last-update-time))
-                                             magit-todos-update)
-                                         (null magit-todos-last-update-time)))))
-     ;; Scan and insert.
-     ;; HACK: I don't like setting a special var here, because it seems like lexically binding a
-     ;; special var should follow down the chain, but it isn't working, so we'll do this.
-     (setq magit-todos-updating t)
-     (setq magit-todos-active-scan (funcall magit-todos-scanner
-                                            :callback #'magit-todos--insert-items
-                                            :magit-status-buffer (current-buffer)
-                                            :directory default-directory
-                                            :depth magit-todos-depth)))
-    (_ ; Caching and cache not expired, or not automatic and not manually updating now
-     (magit-todos--insert-items (current-buffer) magit-todos-item-cache)))
+  (when (or magit-todos-update-remote
+            magit-todos-updating
+            (not (file-remote-p default-directory)))
+    (when magit-todos-active-scan
+      ;; Avoid running multiple scans for a single magit-status buffer.
+      (let ((buffer (process-buffer magit-todos-active-scan)))
+        (when (process-live-p magit-todos-active-scan)
+          (delete-process magit-todos-active-scan))
+        (when (buffer-live-p buffer)
+          (kill-buffer buffer)))
+      (setq magit-todos-active-scan nil))
+    (pcase magit-todos-update
+      ((or 't                           ; Automatic
+           ;; Manual and updating now
+           (and 'nil (guard magit-todos-updating))
+           ;; Caching and cache expired
+           (and (pred integerp) (guard (or magit-todos-updating ; Forced update
+                                           (>= (float-time
+                                                (time-subtract (current-time)
+                                                               magit-todos-last-update-time))
+                                               magit-todos-update)
+                                           (null magit-todos-last-update-time)))))
+       ;; Scan and insert.
+       ;; HACK: I don't like setting a special var here, because it seems like lexically binding a
+       ;; special var should follow down the chain, but it isn't working, so we'll do this.
+       (setq magit-todos-updating t)
+       (setq magit-todos-active-scan (funcall magit-todos-scanner
+                                              :callback #'magit-todos--insert-items
+                                              :magit-status-buffer (current-buffer)
+                                              :directory default-directory
+                                              :depth magit-todos-depth))
+       (magit-todos--maybe-insert-branch-todos 'rescan))
+      (_ ; Caching and cache not expired, or not automatic and not manually updating now
+       (magit-todos--insert-items (current-buffer) magit-todos-item-cache)
+       (magit-todos--maybe-insert-branch-todos 'cached)))))
+
+(defun magit-todos--maybe-insert-branch-todos (&optional type)
+  "Insert branch todos when appropriate.
+If TYPE is `rescan', rescan for items; if `cached', only insert
+cached items, if any; otherwise, use cached items if any, or
+rescan."
   (when (or (eq magit-todos-branch-list t)
             (and (eq magit-todos-branch-list 'branch)
                  (not (equal (or magit-todos-branch-list-merge-base-ref (magit-main-branch))
                              (magit-get-current-branch)))))
-    ;; Insert branch-local items.
-    (magit-todos--scan-with-git-diff :magit-status-buffer (current-buffer)
-                                     :directory default-directory
-                                     :depth magit-todos-depth
-                                     :heading (format "TODOs (branched from %s)"
-                                                      (or magit-todos-branch-list-merge-base-ref (magit-main-branch))))))
+    (if (or (eq 'rescan type)
+            (and (not (eq 'cached type))
+                 (null magit-todos-branch-item-cache)))
+        ;; TODO: Refactor to just return items and then insert separately.
+        (magit-todos--scan-with-git-diff
+         :magit-status-buffer (current-buffer)
+         :directory default-directory
+         :depth magit-todos-depth
+         :heading (format "TODOs (branched from %s)"
+                          (or magit-todos-branch-list-merge-base-ref (magit-main-branch))))
+      ;; Just insert cached items, if any.
+      (let ((magit-todos-section-heading
+             (format "TODOs (branched from %s)"
+                     (or magit-todos-branch-list-merge-base-ref (magit-main-branch)))))
+        (magit-todos--insert-items (current-buffer) magit-todos-branch-item-cache :branch-p t)))))
 
 (cl-defun magit-todos--insert-items (magit-status-buffer items &key branch-p)
   "Insert to-do ITEMS into MAGIT-STATUS-BUFFER.
@@ -980,7 +984,14 @@ This should be called in a process's output buffer from one of
 the async callback functions.  The calling function should
 advance to the next line."
   (let ((case-fold-search magit-todos-ignore-case))
-    (when (re-search-forward regexp (line-end-position) t)
+    (when (condition-case err
+              (re-search-forward regexp (line-end-position) t)
+            (error (if (string-match-p "Stack overflow in regexp matcher" (error-message-string err))
+                       ;; FIXME: This may raise multiple warnings per file.
+                       (let ((filename (buffer-substring (point) (1- (re-search-forward ":")))))
+                         (message "magit-todos: File has lines too long for Emacs to search.  Consider excluding it from scans: %s" filename)
+                         nil)
+                     (signal (car err) (cdr err)))))
       (make-magit-todos-item :filename (or filename
                                            (match-string 8))
                              :line (--when-let (match-string 2)
@@ -1151,7 +1162,7 @@ if the process's buffer has already been deleted."
 
 ;;;; Scanners
 
-(cl-defmacro magit-todos-defscanner (name &key test command results-regexp
+(cl-defmacro magit-todos-defscanner (name &key availablep command results-regexp
                                           (allow-exit-codes '(0))
                                           (directory-form '(f-relative directory default-directory))
                                           (callback (function 'magit-todos--scan-callback)))
@@ -1160,7 +1171,7 @@ if the process's buffer has already been deleted."
 NAME is a string, which may contain spaces.  It is only used for
 descriptive purposes.
 
-TEST is an unquoted sexp which is used to determine whether the
+AVAILABLEP is a predicate which is used to determine whether the
 scanner is usable.  In most cases, it should use
 `executable-find' to look for the scanner command.
 
@@ -1344,7 +1355,7 @@ When SYNC is non-nil, match items are returned."
        (add-to-list 'magit-todos-scanners
                     (list (cons 'name ,name)
                           (cons 'function #',scan-fn-symbol)
-                          (cons 'test ',test))
+                          (cons 'availablep ,availablep))
                     'append))))
 
 ;; NOTE: These scanners handle the max-depth option differently.  git-grep seems to handle it in the
@@ -1357,7 +1368,8 @@ When SYNC is non-nil, match items are returned."
 ;; scanners we'll add one to its value.
 
 (magit-todos-defscanner "rg"
-  :test (executable-find "rg")
+  :availablep (lambda ()
+                (executable-find "rg"))
   :directory-form (if (equal directory default-directory)
                       ;; Prevent leading "./" in filenames.
                       nil
@@ -1377,7 +1389,14 @@ When SYNC is non-nil, match items are returned."
                  extra-args search-regexp-pcre directory))
 
 (magit-todos-defscanner "git grep"
-  :test (string-match "--perl-regexp" (shell-command-to-string "git grep --magit-todos-testing-git-grep"))
+  ;; To allow git-grep to work regardless of whether `default-directory'
+  ;; is in a Git repository, we use "--no-index" (which acts like "grep
+  ;; -r") and "-" (STDIN, so we needn't specify files).  git-grep exits
+  ;; this test with 128 if PCRE is not supported.  We also allow exit
+  ;; code 1, because it means a successful grep run (i.e. "--perl-regexp"
+  ;; is supported) without matches.
+  :availablep (lambda ()
+                (>= 1 (call-process-shell-command "git grep --no-index --quiet --perl-regexp '\\d' -- -")))
   :allow-exit-codes (0 1)
   :command (list "git" "--no-pager" "grep"
                  "--full-name" "--no-color" "-n"
@@ -1398,7 +1417,6 @@ When SYNC is non-nil, match items are returned."
 (magit-todos-defscanner "git diff"
   ;; NOTE: This scanner implements the regexp *searching* in elisp rather than in the
   ;; external process because, unlike "git grep", "git diff" does not support PCRE.
-  :test t
   :command (progn
              ;; Silence byte-compiler warnings about these vars we don't use in this scanner.
              (ignore search-regexp-elisp search-regexp-pcre extra-args directory depth)
@@ -1407,13 +1425,14 @@ When SYNC is non-nil, match items are returned."
                                        shell-command-to-string
                                        string-trim)))
                (unless (string-empty-p merge-base-ref)
-                 (list "git" "--no-pager" "diff" "--no-color" "-U0" merge-base-ref))))
+                 (list "git" "--no-pager" "diff" "--no-ext-diff" "--no-color" "-U0" merge-base-ref))))
   :callback 'magit-todos--git-diff-callback)
 
 (magit-todos-defscanner "find|grep"
   ;; NOTE: The filenames output by find|grep have a leading "./".  I don't expect this scanner to be
   ;; used much, if at all, so I'm not going to go to the trouble to fix this now.
-  :test (string-match "--perl-regexp" (shell-command-to-string "grep --help"))
+  :availablep (lambda ()
+                (string-match "--perl-regexp" (shell-command-to-string "grep --help")))
   :allow-exit-codes (0 1)
   :command (let* ((grep-find-template (progn
                                         (unless grep-find-template
